# include<iostream>
# include<string>
# include<stdio.h>
# include<StringBasics.h>
# include<StringArray.h>
# include <StringHash.h>
# include <algorithm>
# include <vector>
# include <sstream>
# include <iterator>
# include <cmath>
# include <map>
#include <sstream>
#include "Parameters.h"

using namespace std;

#define DELIM '\t'
#define DISTANCE 5


const char * codons[64] = {
	"AAA", "AAC", "AAG", "AAT", "ACA", "ACC", "ACG", "ACT", "AGA", "AGC", "AGG",
	"AGT", "ATA", "ATC", "ATG", "ATT", "CAA", "CAC", "CAG", "CAT", "CCA", "CCC",
	"CCG", "CCT", "CGA", "CGC", "CGG", "CGT", "CTA", "CTC", "CTG", "CTT", "GAA",
	"GAC", "GAG", "GAT", "GCA", "GCC", "GCG", "GCT", "GGA", "GGC", "GGG", "GGT",
	"GTA", "GTC", "GTG", "GTT", "TAA", "TAC", "TAG", "TAT", "TCA", "TCC", "TCG",
	"TCT", "TGA", "TGC", "TGG", "TGT", "TTA", "TTC", "TTG", "TTT" };

const char * codons2aminoAcids [64] = {
	"Lys", "Asn", "Lys", "Asn", "Thr", "Thr", "Thr", "Thr", "Arg", "Ser", "Arg",
	"Ser", "Ile", "Ile", "Met", "Ile", "Gln", "His", "Gln", "His", "Pro", "Pro",
	"Pro", "Pro", "Arg", "Arg", "Arg", "Arg", "Leu", "Leu", "Leu", "Leu", "Glu",
	"Asp", "Glu", "Asp", "Ala", "Ala", "Ala", "Ala", "Gly", "Gly", "Gly", "Gly",
	"Val", "Val", "Val", "Val", "stop", "Tyr", "stop", "Tyr", "Ser", "Ser", "Ser",
	"Ser", "stop", "Cys", "Trp", "Cys", "Leu", "Phe", "Leu", "Phe" };

const char * codons2letters [64] = {
	"K", "N", "K", "N", "T", "T", "T", "T", "R", "S", "R", "S", "I", "I", "M",
	"I", "Q", "H", "Q", "H", "P", "P", "P", "P", "R", "R", "R", "R", "L", "L",
	"L", "L", "E", "D", "E", "D", "A", "A", "A", "A", "G", "G", "G", "G", "V",
	"V", "V", "V", "*", "Y", "*", "Y", "S", "S", "S", "S", "*", "C", "W", "C",
	"L", "F", "L", "F" };

const char * aminoAcids [21] = {
	"Ter", "Ala", "Cys", "Asp", "Glu", "Phe", "Gly", "His", "Ile", "Lys", "Leu",
	"Met", "Asn", "Pro", "Gln", "Arg", "Ser", "Thr", "Val", "Trp", "Tyr" };

const char * aminoAcidLetters [21] = {
	"*", "A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "P", "Q",
	"R", "S", "T", "V", "W", "Y" };


char ReverseSeq (char inword)
{
	char outword;
	if ((inword == 'A') or (inword == 'a')) {outword = 'T';}
	else if ((inword == 'G') or (inword == 'g')) {outword = 'C';}
	else if ((inword == 'T') or (inword == 't')) {outword = 'A';}
	else if ((inword == 'C') or (inword == 'c')) {outword = 'G';}
	else {outword = '-';}
	return outword;
}

char UppercaseSeq (char inword)
{
	char outword;
	if ((inword == 'A') or (inword == 'a')) {outword = 'A';}
	else if ((inword == 'G') or (inword == 'g')) {outword = 'G';}
	else if ((inword == 'T') or (inword == 't')) {outword = 'T';}
	else if ((inword == 'C') or (inword == 'c')) {outword = 'C';}
	else {outword = '-';}
	return outword;
}
		

int main(int argc, char *argv[])
{

		//if (argc != 2) {
		//usage(argv[0]);
		//}
   printf("\n\n##################################################################################################\n\n");
#ifndef VERSION
   printf("vcfCodingSNP 1.5 -- vcf SNP annotating tool\n"
#else
   printf("vcfCodingSNP "VERSION" -- vcf SNP annotating tool\n"
#endif
          "(c) 2010.12 Yanming Li, Goncalo Abecasis\nCommend and (or) suggestions are welcome! Please send to liyanmin@umich.edu.\n\n");
   printf("##################################################################################################\n");


    //String reference = "data/genome1.fa";

    String VCFsnpFile, refGenome("referenceGenomes/genome.V36.fa"), geneFile("geneLists/UCSCknownGene.B36.txt"), outFile("vcfCodingSNP.out.vcf"), logFile("vcfCodingSNP.log");
	int n1splice=8, n2splice=3, nstream=5;

	ParameterList pl;

	BEGIN_LONG_PARAMETERS(parameters)
		LONG_PARAMETER_GROUP("Input Files")
			LONG_STRINGPARAMETER("refgenome", &refGenome)
			LONG_STRINGPARAMETER("snpfile", &VCFsnpFile)
			LONG_STRINGPARAMETER("genefile", &geneFile)
            LONG_INTPARAMETER("n1", &n1splice)
            LONG_INTPARAMETER("n2", &n2splice)
            LONG_INTPARAMETER("ns", &nstream)

            
	LONG_PARAMETER_GROUP("Output Files")
			LONG_STRINGPARAMETER("outfile", &outFile)
			LONG_PARAMETER("log", &logFile)
	
	END_LONG_PARAMETERS();
	
	pl.Add(new HiddenString('s', "snpfile", VCFsnpFile));
	pl.Add(new HiddenString('r', "refgenome", refGenome));
	pl.Add(new HiddenString('g', "genefile", geneFile));
	pl.Add(new HiddenString('o', "outfile", outFile));
	pl.Add(new HiddenString('l', "logfile", logFile));    
	pl.Add(new HiddenInteger('i', "n1", n1splice));    
	pl.Add(new HiddenInteger('e', "n1", n2splice));    
	//pl.Add(new HiddenInteger('s', "n", nstream));    


	
	pl.Add(new LongParameters("Availabe Options",parameters));
	
	//usage(argc, argv);
	pl.Read(argc, argv);
	pl.Status(); 

    
		FILE *refgenome=fopen(refGenome, "rt");
        FILE *snpfile = fopen(VCFsnpFile,"rt");
			if(snpfile==NULL){printf("Can't open VCF snpfile!!!\n"); exit(1);}   
        FILE *genefile = fopen(geneFile,"rt");
	    FILE *outfile = fopen(outFile,"wt");
		FILE *logfile = fopen(logFile, "wt");
		if(n1splice<3){printf("Invalid n1 value!!! n1 must be an integer greater or equals to 3.\n");}
		if(n2splice<0){printf("Invalid n2 value!!! n2 must be a non-negative integer.\n");}
		if(nstream<=0){printf("Invalid n1 value!!! ns must be a positive integer.\n");}
		    

	   // FILE *file=fopen("/data/local/ref/GATK/human_g1k_v37.fasta", "rt");
	   // FILE *file=fopen("genome1.fa", "rt");
		//FILE *file=fopen("../hg18_knownGeneMrna.fa", "rt");

    
		//if (snpfile == NULL || genefile == NULL || outfile == NULL || file == NULL) { printf("can't open input file !\n"); exit(0);}


	//**********************************************//
	//* Read reference genome file into the memory *//
	// *********************************************//
    
    
		String input; 
        String chrseq[25];
        //char allele;
		int c=-1;
 
        
		while(!feof(refgenome))
		{
			input.ReadLine(refgenome);

			if (input[0] == '>' && input[1]!= 'M' && input[1]!='N' && input[1]!='G')
			{
				printf("Reading chromosome %s...\n",
						(const char *) input);  
                c++;
				continue;
			}
			chrseq[c]+=input;
		}
		
		

	//**********************************************//
	//* Read SNP file and genefile into the memory *//
	//**********************************************//
        
        long int snpmapchr;
        long int snpmappos;
		long int genemapchr;
		long int genemapstart;

		int chr;
		int pos;
		//int geneindex;
		String genestringchr;
		int endgenestringchr;
		int start_of_underbar;
		int enddigit;
		String genesubchr;
		//int genechr;
		//int genestart;
		int geneend;

		StringArray tokens;
        String buffer;
        
        StringArray tokens1;
        String buffer1;
        
        printf("start mapping\n"); 
         
		printf("mapping snp file... ...");

       ////   snp map ////


        String refallele, altallele;
        
        map <long int, String> snpmap;
        
       long int index = 0;
       
       bool prefixchr = false;
        
        while(!feof(snpfile))
        {   index ++;

			buffer.ReadLine(snpfile); if (buffer.IsEmpty()) continue;

			if (buffer[0]=='#') {fprintf(outfile, "%s\n", (const char*) buffer);}
			else
			{
			tokens.AddTokens(buffer,DELIM);
			
			
			// check if the chrStr is of the format chr1, chr2,...
            if (tokens[0].Left(3) == "chr" || tokens[0].Left(3) == "CHR") {
                tokens[0].LeftClip(3);
                prefixchr = true;
            }


			String chrStr = tokens[0][0];
			if (chrStr=="X") { 
			snpmapchr =23;}
			else if (chrStr=="Y" ){
			snpmapchr=24;
			} else if ( (chrStr=="G") || (chrStr=="MT" ))
			{fprintf(outfile, "%s\n", (const char*) buffer); snpmapchr=99;}
	else   
            { snpmapchr = tokens[0].AsInteger();}
            snpmappos = tokens[1].AsInteger();


			//snpmap.insert(pair<long int, String>(chr*1e9+pos, buffer));
			String refallele = tokens[3];
			String altallele = tokens[4];


			String beforealtallele = tokens[0]+"\t"+ tokens[1]+"\t"+ tokens[2]+"\t"+ tokens[3];
			String afteraltallele = tokens[5];

			for (int i=6; i<=tokens.Length()-1; i++)
			{afteraltallele = afteraltallele+"\t"+ tokens[i];}
			

			StringArray altAlleleArr;
			altAlleleArr.AddTokens(altallele, ",");
            
			
			for (long int i=1; i<=altAlleleArr.Length(); i++)
			{
			String buffer1 = beforealtallele + "\t" + altAlleleArr[i-1] + "\t" +afteraltallele;
            snpmap.insert(pair<long int, String>(snpmapchr*1e12+snpmappos*1000+index*100+i,buffer1));
			//printf("%s\n", (const char*) buffer1);
			buffer1.Clear();
			}
			
			altAlleleArr.Clear();
			//Bufferline.Clear();
			

			tokens.Clear();
			buffer.Clear();
			}
         }   
        
         printf("DONE! snp mapsize = %d\n", (int)snpmap.size());
         
         map <long int, String>::iterator it, lastit; 
              
         
          
          //// read gene list file ////

		long int gindex=0; 
		
		printf("mapping gene file... ...");
                 
        map <long int, String> genemap;
                        
        while(!feof(genefile))
        {              
			gindex++;

			buffer.ReadLine(genefile); if (buffer.IsEmpty()) continue;
	
			if (buffer[0]=='#') {continue;}

			tokens.AddTokens(buffer,DELIM);     
		    //geneindex =tokens[0].AsInteger();

			genestringchr = tokens[1];
			endgenestringchr =genestringchr.Length();
			start_of_underbar= genestringchr.Find("_");
			enddigit=min(endgenestringchr, abs(10000*min(0,start_of_underbar))+start_of_underbar);
			String genesubchr = genestringchr.SubStr(3, enddigit-3);
			if (genesubchr=="X") { 
			genemapchr =23;}
			else if (genesubchr=="Y" ){
			genemapchr=24;}
			else{genemapchr = genesubchr.AsInteger();}
			genemapstart = tokens[3].AsInteger();
			geneend = tokens[4].AsInteger();

 
            genemap.insert(pair<long int, String>(genemapchr*1e14+genemapstart*1e5+gindex,buffer));

			tokens.Clear();
			buffer.Clear();
          
         }   
        
         printf("DONE! gene mapsize = %d\n", (int)genemap.size());
         

		 map<long int, String>::iterator itg, lastitg;
             

        //***************************//
        //*      end test map      *//     
        //**************************//

		printf("start annotating... ...\n"); 
       
		itg =genemap.begin();            
		lastitg=itg;

        fprintf(logfile, "chr\tpos\tref\talt\tucsc_name\tgenestrend\tgenestart\tgeneend\tref_codon\tref_AA\talt_codon\talt_AA\tcodon_start\tcodon_end\tgenesymbol\tcodonCount\ttype\n");

			it=snpmap.begin();
		//	printf("%s\n", (const char*) (*it).second);

		for (it=snpmap.begin(); it!=snpmap.end(); it++)
		{	
            
				tokens.AddTokens((*it).second, DELIM);
				String chrStr = tokens[0][0];
			if (chrStr=="X") { 
			chr =23;}
			else if (chrStr=="Y" ){
			chr=24;
			} else if ( (chrStr=="G") || (chrStr=="MT" ))
			{chr=99;}
				else    
	            { chr = tokens[0].AsInteger();}
				pos = tokens[1].AsInteger();
				refallele = tokens[3][0];
				String allelestr=tokens[4];
				char allele = tokens[4][0];
				String Info =tokens[7];

		
				String beforeInfo = tokens[0]+"\t"+tokens[1]+"\t"+tokens[2]+"\t"+tokens[3]+"\t"+tokens[4]+"\t"+tokens[5]+"\t"+tokens[6];
				String afterInfo = tokens[8];

				for (int i=9; i<=tokens.Length()-1; i++)
				{afterInfo = afterInfo+"\t"+ tokens[i];}

				//StringHash typeArr;

        if (prefixchr == false)
            fprintf(outfile, "%s\t%s", (const char*)beforeInfo, (const char*)Info);
        else
            fprintf(outfile, "chr%s\t%s", (const char*)beforeInfo, (const char*)Info);


				//printf("%s\t%s\n", (const char*)beforeInfo, (const char*)Info);

				itg = lastitg;	

				int flagInGene=0;

        while (itg!= genemap.end() && allelestr!="." && allelestr!="NA")
			{   
            
				tokens1.AddTokens((*itg).second,DELIM);
				
				//int counter = tokens1[0];
				String genename = tokens1[0];
				String genesymbol = tokens1[10];
				
				String genestringchr = tokens1[1];
				int endgenestringchr =genestringchr.Length();
				int start_of_underbar= genestringchr.Find("_");
				int enddigit=min(endgenestringchr, abs(10000*min(0,start_of_underbar))+start_of_underbar);
				String genesubchr = genestringchr.SubStr(3, enddigit-3);
				int genechr;
				if (genesubchr=="X") { 
				genechr =23;}
				else if (genesubchr=="Y" ){
				genechr=24;}
				else{genechr = genesubchr.AsInteger();}
				//int genechr = genesubchr.AsInteger();
				
				String genestr = tokens1[2];
				int genestart = tokens1[3].AsInteger();
				int geneend = tokens1[4].AsInteger();
				int codestart = tokens1[5].AsInteger();
				int codeend = tokens1[6].AsInteger();
				int exonct = tokens1[7].AsInteger();
				String exonstartstr = tokens1[8];
				String exonendstr = tokens1[9];
				
				StringArray exonStartArr;
				StringArray exonEndArr;
				exonStartArr.AddTokens(exonstartstr, ",");
				exonEndArr.AddTokens(exonendstr,",");
				


		        int codepos;
				String snpcodon, seqcodon;
		        int codonstartpos, codonendpos;
				String type;
				int tmpexonstartpos, tmpexonendpos;
 
				
				
if (genestart < geneend && chr == genechr &&  pos >  genestart && pos <= geneend && flagInGene ==0 ){ 
	
	lastitg =itg; 
	
	flagInGene=1; }

	//See if UTR site or not

			if (chr == genechr && exonStartArr[0].AsInteger()<pos && pos<= codestart && genestr=="+") {
			type="5'UTR"; 
			//typeArr.Add(type);
			fprintf(outfile, "%s%s=%s%s%s%s%s%s%s", (const char*)";", (const char*)type, (const char*)genesymbol, (const char*)"(", (const char*)genename, (const char*)")", (const char*)"[", (const char*)genestr,(const char*)"]");
    		fprintf(logfile, "%s\t%d\t%c\t%c\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", (const char*)genestringchr, pos, chrseq[chr-1][pos-1],allele, (const char*)genename, (const char*)genestr, genestart,geneend, (const char*)".", (const char*) ".", (const char*)".", (const char*) ".",  (const char*) ".", (const char*) ".", (const char*)genesymbol, (const char*) ".", (const char*)type );

			itg ++;  // cw add
				exonStartArr.Clear();
				exonEndArr.Clear();
		        tokens1.Clear();
		        buffer1.Clear();
				continue;
			}

			else if (chr == genechr && exonStartArr[0].AsInteger()<pos && pos<= codestart && genestr=="-") {
			type="3'UTR";
			//typeArr.Add(type);
			fprintf(outfile, "%s%s=%s%s%s%s%s%s%s", (const char*)";", (const char*)type, (const char*)genesymbol, (const char*)"(", (const char*)genename, (const char*)")", (const char*)"[", (const char*)genestr,(const char*)"]");
		    fprintf(logfile, "%s\t%d\t%c\t%c\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", (const char*)genestringchr, pos, chrseq[chr-1][pos-1],allele, (const char*)genename, (const char*)genestr, genestart,geneend, (const char*)".", (const char*) ".", (const char*)".", (const char*) ".",  (const char*) ".", (const char*) ".", (const char*)genesymbol, (const char*) ".", (const char*)type );

			itg ++;  // cw add
				exonStartArr.Clear();
				exonEndArr.Clear();
		        tokens1.Clear();
		        buffer1.Clear();
				continue;
			}

			else if (chr == genechr && codeend<pos && pos<= exonEndArr[exonEndArr.Length()-1].AsInteger() && genestr=="+"){ 
			type="3'UTR";
			//typeArr.Add(type);
			fprintf(outfile, "%s%s=%s%s%s%s%s%s%s", (const char*)";", (const char*)type, (const char*)genesymbol, (const char*)"(", (const char*)genename, (const char*)")", (const char*)"[", (const char*)genestr,(const char*)"]");
		    fprintf(logfile, "%s\t%d\t%c\t%c\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", (const char*)genestringchr, pos, chrseq[chr-1][pos-1],allele, (const char*)genename, (const char*)genestr, genestart,geneend, (const char*)".", (const char*) ".", (const char*)".", (const char*) ".",  (const char*) ".", (const char*) ".", (const char*)genesymbol, (const char*) ".", (const char*)type );

			itg ++;
			exonStartArr.Clear();
			exonEndArr.Clear();
		    tokens1.Clear();
		    buffer1.Clear();
			continue;
			}

			else if (chr == genechr && codeend<pos && pos<= exonEndArr[exonEndArr.Length()-1].AsInteger() && genestr=="-"){
			type="5'UTR";
			//typeArr.Add(type);
			fprintf(outfile, "%s%s=%s%s%s%s%s%s%s", (const char*)";", (const char*)type, (const char*)genesymbol, (const char*)"(", (const char*)genename, (const char*)")", (const char*)"[", (const char*)genestr,(const char*)"]");
		    fprintf(logfile, "%s\t%d\t%c\t%c\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", (const char*)genestringchr, pos, chrseq[chr-1][pos-1],allele, (const char*)genename, (const char*)genestr, genestart,geneend, (const char*)".", (const char*) ".", (const char*)".", (const char*) ".",  (const char*) ".", (const char*) ".", (const char*)genesymbol, (const char*) ".", (const char*)type );

			itg ++;
				exonStartArr.Clear();
				exonEndArr.Clear();
		        tokens1.Clear();
		        buffer1.Clear();
				continue;
			}


			else if (codestart < codeend && chr == genechr &&  pos >  codestart && pos <= codeend ) 
        	{                
				if (exonStartArr.Length() != exonEndArr.Length() || exonStartArr.Length() != exonct) {
			    printf("Error: Exon count is not consistent with the length of the exon start vector!!!\n");
			    exit(0);
				}

			//find the exon number for the coding start site
            //find the exon number for the coding end site
   		    //find the exon number for the SNP pos

			int csexonpos = 0;
			int ceexonpos = 0;
		 	int snpexonpos = 0;
			String	type0="";

			for (int i = 1; i <= exonct; i++) {
			if ( codestart >= exonStartArr[i-1].AsInteger() && codestart < exonEndArr[i-1].AsInteger() ) {
				csexonpos = i;
				//break;
				}

			if ( codeend-1 >= exonStartArr[i-1].AsInteger() && codeend-1 < exonEndArr[i-1].AsInteger() ) {
		                ceexonpos = i;
			              //break;
				}

			if ( pos-1 >= exonStartArr[i-1].AsInteger() && pos-1 < exonEndArr[i-1].AsInteger() ) {
				snpexonpos = i;
				//break;
				}
			if ( ((pos<=exonStartArr[i-1].AsInteger()+n2splice) && (pos > exonStartArr[i-1].AsInteger())) || ((exonEndArr[i-1].AsInteger()-n2splice< pos) && (pos<= exonEndArr[i-1].AsInteger()))) {type0="(SPLICE_SITE)";}

			}

		  	if ( csexonpos == 0 ) {
			    printf("Error: coding start position is not right!!!\n");
			    exit(0);
			}
                  
            if ( ceexonpos == 0 ) {
				printf("Error: coding end position is not right!!!\n");
				exit(0);
	            }

		  
			if ( snpexonpos == 0 ) {
				type="INTRONIC";
				if ((pos <= exonEndArr[0].AsInteger()+2) || (exonStartArr[exonct-1].AsInteger()-1 <= pos)) {type="ESSENTIAL_SPLICE_SITE";}
				else if (((pos>exonEndArr[0].AsInteger()+2) && (pos <= exonEndArr[0].AsInteger()+n1splice)) || ((pos>=exonStartArr[exonct-1].AsInteger()-n1splice+1) && (pos<exonStartArr[exonct-1].AsInteger()-1))) {type="SPLICE_SITE";}
				for(int i=2; i <= exonct-1; i++) {
				if ( ((pos==exonStartArr[i-1].AsInteger()) || (pos==exonStartArr[i-1].AsInteger()-1)) || ((pos==exonEndArr[i-1].AsInteger()+1) && (pos== exonEndArr[i-1].AsInteger()+2))) {type="ESSENTIAL_SPLICE_SITE";}
				else if ( ((pos< exonStartArr[i-1].AsInteger()-1) && (exonStartArr[i-1].AsInteger()-n1splice+1<=pos)) || ((exonEndArr[i-1].AsInteger()+2< pos) && (pos<= exonEndArr[i-1].AsInteger()+n1splice))) {type="SPLICE_SITE";}
				}	
	
				fprintf(outfile, "%s%s=%s%s%s%s%s%s%s", (const char*)";", (const char*)type, (const char*)genesymbol, (const char*)"(", (const char*)genename, (const char*)")", (const char*)"[", (const char*)genestr,(const char*)"]");
			    fprintf(logfile, "%s\t%d\t%c\t%c\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", (const char*)genestringchr, pos, chrseq[chr-1][pos-1],allele, (const char*)genename, (const char*)genestr, genestart,geneend, (const char*)".", (const char*) ".", (const char*)".", (const char*) ".",  (const char*) ".", (const char*) ".", (const char*)genesymbol, (const char*) ".", (const char*)type );


				//typeArr.Add(type);
				itg ++;  // cw add
				exonStartArr.Clear();
				exonEndArr.Clear();
		        tokens1.Clear();
		        buffer1.Clear();
				continue;
			}

			/*for(int i=1; i <= exonct; i++) {
					if ( ((pos<=exonStartArr[i-1].AsInteger()+n2splice) && (pos > exonStartArr[i-1].AsInteger())) || ((exonEndArr[i-1].AsInteger()-n2splice< pos) && (pos<= exonEndArr[i-1].AsInteger()))) {type="(SPLICE_SITE)";}
					else {type="";}
					}*/

	
			// Calculate the number of bases between the coding start site and the coding end site on mRNA to double check. 		  
			int csexonposBak = csexonpos;
			int numBase = 0;
			if (csexonposBak == ceexonpos) {
				numBase = codeend - codestart;
				}
			if (ceexonpos > csexonposBak) {
				numBase = exonEndArr[csexonposBak-1].AsInteger() - codestart;
				csexonposBak++;
				while (ceexonpos > csexonposBak) {
					numBase = numBase + exonEndArr[csexonposBak-1].AsInteger() - exonStartArr[csexonposBak-1].AsInteger();
					csexonposBak++;
				}
		  		numBase = numBase + codeend - exonStartArr[csexonposBak-1].AsInteger();
		 	}

			int totalcodonCount;
			int codonmode;
			codonmode = numBase % 3;
			if (codonmode != 0) {
			printf ("Error: The total number of bases between coding start and coding end are not right! genename=%s and base # btw cdstart and cdend=%d\n",(const char*) genename, numBase);
		  totalcodonCount =0;}
			else {
				totalcodonCount = numBase/3;
			}


		// Calculate the physical distance between the SNP and the coding start site. 
		// Example: phyDistance = 0 if the SNP and the coding start site lie at the same position.

			csexonposBak = csexonpos; 
		 	int phyDistance = 0;
		  	if (snpexonpos == csexonposBak) {
		    	phyDistance = pos - (codestart+1);
		  	}
			if (snpexonpos > csexonposBak) {
				phyDistance = exonEndArr[csexonposBak-1].AsInteger() - (codestart+1);
			    csexonposBak++;
			    while (snpexonpos > csexonposBak) {
					phyDistance = phyDistance + exonEndArr[csexonposBak-1].AsInteger() - exonStartArr[csexonposBak-1].AsInteger();
					csexonposBak++;
			    }
				phyDistance = phyDistance + pos - exonStartArr[csexonposBak-1].AsInteger();
			}
		  
			int codonCount;
			codonCount = int(floor(double(phyDistance)/3.0) + 1);
			codepos = phyDistance % 3;
			codonstartpos = pos - codepos;  // start position of the 3-base codon(always based on the + strand)
			codonendpos = codonstartpos + 2; // end position of the 3-base codon (always based on the + strand)

			if (genestr=="+")
			 {	   	
					if (codepos == 0) {
						if (pos==exonEndArr[snpexonpos-1].AsInteger()){
							tmpexonstartpos= exonStartArr[snpexonpos].AsInteger();
							snpcodon = (String)UppercaseSeq(allele) + (String)UppercaseSeq(chrseq[chr-1][tmpexonstartpos]) + (String)UppercaseSeq(chrseq[chr-1][tmpexonstartpos+1]); 
							seqcodon = (String)UppercaseSeq(chrseq[chr-1][pos-1]) + (String)UppercaseSeq(chrseq[chr-1][tmpexonstartpos]) + (String)UppercaseSeq(chrseq[chr-1][tmpexonstartpos+1]); 
							}

						else if (pos==exonEndArr[snpexonpos-1].AsInteger()-1){
							tmpexonstartpos=exonStartArr[snpexonpos].AsInteger();
							snpcodon = (String)UppercaseSeq(allele) + (String)UppercaseSeq(chrseq[chr-1][pos]) + (String)UppercaseSeq(chrseq[chr-1][tmpexonstartpos]); 
							seqcodon = (String)UppercaseSeq(chrseq[chr-1][pos-1]) + (String)UppercaseSeq(chrseq[chr-1][pos]) + (String)UppercaseSeq(chrseq[chr-1][tmpexonstartpos]); 
							}

						else{
							snpcodon = (String)UppercaseSeq(allele) + (String)UppercaseSeq(chrseq[chr-1][pos]) + (String)UppercaseSeq(chrseq[chr-1][pos+1]); 
							seqcodon =  (String)UppercaseSeq(chrseq[chr-1][pos-1]) + (String)UppercaseSeq(chrseq[chr-1][pos]) + (String)UppercaseSeq(chrseq[chr-1][pos+1]); 
							}
						}


					else if (codepos == 1) {
						if (pos==exonEndArr[snpexonpos-1].AsInteger()){
							tmpexonstartpos=exonStartArr[snpexonpos].AsInteger();
							snpcodon =  (String)UppercaseSeq(chrseq[chr-1][pos-2]) + (String)UppercaseSeq(allele) + (String)UppercaseSeq(chrseq[chr-1][tmpexonstartpos]); 
							seqcodon =  (String)UppercaseSeq(chrseq[chr-1][pos-2]) + (String)UppercaseSeq(chrseq[chr-1][pos-1]) + (String)UppercaseSeq(chrseq[chr-1][tmpexonstartpos]);
							}

						else if (pos==(exonStartArr[snpexonpos-1].AsInteger()+1)){
							tmpexonendpos=exonEndArr[snpexonpos-2].AsInteger();
							snpcodon =  (String)UppercaseSeq(chrseq[chr-1][tmpexonendpos-1]) + (String)UppercaseSeq(allele) + (String)UppercaseSeq(chrseq[chr-1][pos]); 
							seqcodon =  (String)UppercaseSeq(chrseq[chr-1][tmpexonendpos-1]) + (String)UppercaseSeq(chrseq[chr-1][pos-1]) + (String)UppercaseSeq(chrseq[chr-1][pos]);
							}

						else {
							snpcodon =  (String)UppercaseSeq(chrseq[chr-1][pos-2]) + (String)UppercaseSeq(allele) + (String)UppercaseSeq(chrseq[chr-1][pos]); 
							seqcodon =  (String)UppercaseSeq(chrseq[chr-1][pos-2]) + (String)UppercaseSeq(chrseq[chr-1][pos-1]) + (String)UppercaseSeq(chrseq[chr-1][pos]);
							}
						}


					else if (codepos == 2){ 
						if (pos==exonStartArr[snpexonpos-1].AsInteger()+1){
							tmpexonendpos=exonEndArr[snpexonpos-2].AsInteger();
							snpcodon =   (String)UppercaseSeq(chrseq[chr-1][tmpexonendpos-2]) +  (String)UppercaseSeq(chrseq[chr-1][tmpexonendpos-1]) + (String)UppercaseSeq(allele);
							seqcodon =  (String)UppercaseSeq(chrseq[chr-1][tmpexonendpos-2]) +  (String)UppercaseSeq(chrseq[chr-1][tmpexonendpos-1]) + (String)UppercaseSeq(chrseq[chr-1][pos-1]);
							}

						else if (pos==exonStartArr[snpexonpos-1].AsInteger()+2){
							tmpexonendpos=exonEndArr[snpexonpos-2].AsInteger();
							snpcodon =  (String)UppercaseSeq(chrseq[chr-1][tmpexonendpos-1]) + (String)UppercaseSeq(chrseq[chr-1][pos-2]) + (String)UppercaseSeq(allele);
							seqcodon =  (String)UppercaseSeq(chrseq[chr-1][tmpexonendpos-1]) + (String)UppercaseSeq(chrseq[chr-1][pos-2]) + (String)UppercaseSeq(chrseq[chr-1][pos-1]);
							}

						else {
							snpcodon =  (String)UppercaseSeq(chrseq[chr-1][pos-3]) + (String)UppercaseSeq(chrseq[chr-1][pos-2]) + (String)UppercaseSeq(allele);
							seqcodon =  (String)UppercaseSeq(chrseq[chr-1][pos-3]) + (String)UppercaseSeq(chrseq[chr-1][pos-2]) + (String)UppercaseSeq(chrseq[chr-1][pos-1]);
							}
						}
			 }

			else if (genestr=="-"){	   
				codonCount = totalcodonCount - codonCount + 1;
				codepos = 2 - codepos; // Now this is based on the coding end position on the "+" strand (i.e. the real coding start site)
					if (codepos ==0){
							if (pos==exonStartArr[snpexonpos-1].AsInteger()+1){
							tmpexonendpos= exonEndArr[snpexonpos-2].AsInteger();
							snpcodon = (String)ReverseSeq(allele) + (String)ReverseSeq(chrseq[chr-1][tmpexonendpos-1]) + (String)ReverseSeq(chrseq[chr-1][tmpexonendpos-2]);
							seqcodon = (String)ReverseSeq(chrseq[chr-1][pos-1]) + (String)ReverseSeq(chrseq[chr-1][tmpexonendpos-1]) + (String)ReverseSeq(chrseq[chr-1][tmpexonendpos-2]);
							}

							else if (pos==exonStartArr[snpexonpos-1].AsInteger()+2){
							tmpexonendpos= exonEndArr[snpexonpos-2].AsInteger();
							snpcodon = (String)ReverseSeq(allele) + (String)ReverseSeq(chrseq[chr-1][pos-2]) + (String)ReverseSeq(chrseq[chr-1][tmpexonendpos-1]);
							seqcodon = (String)ReverseSeq(chrseq[chr-1][pos-1]) + (String)ReverseSeq(chrseq[chr-1][pos-2]) + (String)ReverseSeq(chrseq[chr-1][tmpexonendpos-1]);
							}

							else {
							snpcodon = (String)ReverseSeq(allele) + (String)ReverseSeq(chrseq[chr-1][pos-2]) + (String)ReverseSeq(chrseq[chr-1][pos-3]);
							seqcodon = (String)ReverseSeq(chrseq[chr-1][pos-1]) + (String)ReverseSeq(chrseq[chr-1][pos-2]) + (String)ReverseSeq(chrseq[chr-1][pos-3]);
							}
						}
							
				    else if (codepos ==1){
							if (pos==exonStartArr[snpexonpos-1].AsInteger()+1){
							tmpexonendpos= exonEndArr[snpexonpos-2].AsInteger();
							snpcodon = (String)ReverseSeq(chrseq[chr-1][pos]) + (String)ReverseSeq(allele) + (String)ReverseSeq(chrseq[chr-1][tmpexonendpos-1]);
							seqcodon =  (String)ReverseSeq(chrseq[chr-1][pos]) + (String)ReverseSeq(chrseq[chr-1][pos-1]) + (String)ReverseSeq(chrseq[chr-1][tmpexonendpos-1]);
							}

							else if (pos==exonEndArr[snpexonpos-1].AsInteger()){
							tmpexonstartpos=exonStartArr[snpexonpos].AsInteger();
							snpcodon = (String)ReverseSeq(chrseq[chr-1][tmpexonstartpos]) + (String)ReverseSeq(allele) + (String)ReverseSeq(chrseq[chr-1][pos-2]);
						    seqcodon =  (String)ReverseSeq(chrseq[chr-1][tmpexonstartpos]) + (String)ReverseSeq(chrseq[chr-1][pos-1]) + (String)ReverseSeq(chrseq[chr-1][pos-2]);
							}

						 	else {
							snpcodon = (String)ReverseSeq(chrseq[chr-1][pos]) + (String)ReverseSeq(allele) + (String)ReverseSeq(chrseq[chr-1][pos-2]);
						    seqcodon =  (String)ReverseSeq(chrseq[chr-1][pos]) + (String)ReverseSeq(chrseq[chr-1][pos-1]) + (String)ReverseSeq(chrseq[chr-1][pos-2]);
							} 
						}  
					else if (codepos==2) {
							if (pos==exonEndArr[snpexonpos-1].AsInteger()){
							tmpexonstartpos=exonStartArr[snpexonpos].AsInteger();
							snpcodon = (String)ReverseSeq(chrseq[chr-1][tmpexonstartpos+1]) + (String)ReverseSeq(chrseq[chr-1][tmpexonstartpos]) + (String)ReverseSeq(allele);
							seqcodon =  (String)ReverseSeq(chrseq[chr-1][tmpexonstartpos+1]) + (String)ReverseSeq(chrseq[chr-1][tmpexonstartpos]) + (String)ReverseSeq(chrseq[chr-1][pos-1]);
							}

							else if (pos==exonEndArr[snpexonpos-1].AsInteger()-1){
							tmpexonstartpos=exonStartArr[snpexonpos].AsInteger();
							snpcodon =  (String)ReverseSeq(chrseq[chr-1][tmpexonstartpos]) + (String)ReverseSeq(chrseq[chr-1][pos]) + (String)ReverseSeq(allele);
							seqcodon =  (String)ReverseSeq(chrseq[chr-1][tmpexonstartpos]) + (String)ReverseSeq(chrseq[chr-1][pos]) + (String)ReverseSeq(chrseq[chr-1][pos-1]);}

							else {
							snpcodon =  (String)ReverseSeq(chrseq[chr-1][pos+1]) + (String)ReverseSeq(chrseq[chr-1][pos]) + (String)ReverseSeq(allele);
							seqcodon =  (String)ReverseSeq(chrseq[chr-1][pos+1]) + (String)ReverseSeq(chrseq[chr-1][pos]) + (String)ReverseSeq(chrseq[chr-1][pos-1]);
							}
						}
			 	}
		
			else {
				printf("Exon strend is not valid!!!\n");
				exit(0);
				}
				 
			int snpindex = 0; 
			int seqindex = 1;
            String tmpcodon;
            for (int i = 0; i < 64; i ++) { 
		    	tmpcodon = tmpcodon.Copy(codons[i]);
				if (tmpcodon == seqcodon) seqindex = i;
				if (tmpcodon == snpcodon) snpindex = i;
			} 

			//string codonCountstr;
			//char ccstr[10];
			//sprintf(ccstr, "%d", i);
			//codonCountstr+=ccstr;
                   
                                                        
			if (String(codons2aminoAcids[seqindex]) == String(codons2aminoAcids[snpindex])){
				type = "SYNONYMOUS_CODING"+type0;
				}

			else if(String(codons2aminoAcids[snpindex])=="stop"){
				type ="STOP_GAINED"+type0;
				}

			else if(String(codons2aminoAcids[seqindex])=="stop"){
				type ="STOP_LOST"+type0;
				}

			else {
				type ="NON_SYNONYMOUS_CODING"+type0;
				} 


			fprintf(outfile, "%s%s=%s%s%s%s%s%s%d%s%s%s%s", (const char*)";", (const char*)type, (const char*)genesymbol, (const char*)"(", (const char*)genename, (const char*)")", (const char*)":", (const char*) codons2aminoAcids[seqindex], codonCount, (const char*) codons2aminoAcids[snpindex], (const char*)"[", (const char*)genestr,(const char*)"]");
            fprintf(logfile, "%s\t%d\t%c\t%c\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t%d\t%d\t%s\t%d\t%s\n", (const char*)genestringchr, pos, chrseq[chr-1][pos-1],allele,(const char*)genename, (const char*)genestr, genestart,geneend, (const char*)seqcodon, codons2aminoAcids[seqindex], (const char*)snpcodon, codons2aminoAcids[snpindex], codonstartpos, codonendpos, (const char*)genesymbol, codonCount, (const char*)type);

                                     
                itg ++;
				//typeArr.Add(type);       
		}   // end of if (codestart < codeend && chr == genechr &&  pos >  codestart && pos <= codeend)
           
//	}	// end of if (genestart < geneend && chr == genechr &&  pos >  genestart && pos <= geneend)
               
		else if (chr < genechr) { exonStartArr.Clear(); exonEndArr.Clear(); tokens1.Clear(); buffer1.Clear(); break; }

		else if (chr == genechr && pos < genestart) { 
			if (genestart-1000*nstream<pos){ 
			type="UPSTREAM";
			fprintf(outfile, "%s%s=%s%s%s%s%s%s%s", (const char*)";", (const char*)type, (const char*)genesymbol, (const char*)"(", (const char*)genename, (const char*)")", (const char*)"[", (const char*)genestr,(const char*)"]");
			 fprintf(logfile, "%s\t%d\t%c\t%c\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", (const char*)genestringchr, pos, chrseq[chr-1][pos-1],allele, (const char*)genename, (const char*)genestr, genestart,geneend, (const char*)".", (const char*) ".", (const char*)".", (const char*) ".",  (const char*) ".", (const char*) ".", (const char*)genesymbol, (const char*) ".", (const char*)type );
			}			
			exonStartArr.Clear(); exonEndArr.Clear(); tokens1.Clear(); buffer1.Clear(); break;}

		else if (chr == genechr && pos > geneend) {   
			if (pos<=genestart+1000*nstream){ 
			type="DOWNSTREAM";
			fprintf(outfile, "%s%s=%s%s%s%s%s%s%s", (const char*)";", (const char*)type, (const char*)genesymbol, (const char*)"(", (const char*)genename, (const char*)")", (const char*)"[", (const char*)genestr,(const char*)"]");
			fprintf(logfile, "%s\t%d\t%c\t%c\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n", (const char*)genestringchr, pos, chrseq[chr-1][pos-1],allele, (const char*)genename, (const char*)genestr, genestart,geneend, (const char*)".", (const char*) ".", (const char*)".", (const char*) ".",  (const char*) ".", (const char*) ".", (const char*)genesymbol, (const char*) ".", (const char*)type );}
		exonStartArr.Clear(); exonEndArr.Clear(); tokens1.Clear(); buffer1.Clear(); /*lastitg = itg;*/ itg ++; continue;}

		else {  exonStartArr.Clear(); exonEndArr.Clear(); tokens1.Clear(); buffer1.Clear(); itg ++; continue; } 
                              
		exonStartArr.Clear();
		exonEndArr.Clear();
        tokens1.Clear();
        buffer1.Clear();                                                           
                							
		}  // end of while

		//fprintf(outfile, "%s\t%s", (const char*)beforeInfo, (const char*)Info);

            //for (int k = 0; k <typeArr.Capacity(); k++)
            //{if (typeArr.SlotInUse(k))
            //fprintf(outfile, ";%s",  (const char*)typeArr[k]);}

		fprintf(outfile, "\t%s\n", (const char*) afterInfo);
		
           
		tokens.Clear();
		buffer.Clear();
          
		} //end of for
		
		printf("DONE! Complete annotating!!!\n");
                      
        fclose(refgenome);
        fclose(snpfile);
		fclose(genefile);
		fclose(outfile);
		fclose(logfile);

		return (1);
        
        
}

